"""
Utility functions to download data from backend data sources so it can be operated on
locally.

When downloading from an EDL-token aware data source, this module uses EDL shared /
federated token authentication.

This module relies on the harmony_service_lib.util.config and its environment variables to be
set for correct operation. See that module and the project README for details.
"""

from functools import lru_cache
import json
from time import sleep
from urllib.parse import urlparse, parse_qs, urlencode, urlunparse
import datetime
import sys
import os
import re
from harmony_service_lib.earthdata import EarthdataAuth, EarthdataSession
from harmony_service_lib.exceptions import ServerException, ForbiddenException
from harmony_service_lib.logging import build_logger

# Timeout in seconds.  Per requests docs, this is not a time limit on
# the entire response download; rather, an exception is raised if the
# server has not issued a response for timeout seconds (more
# precisely, if no bytes have been received on the underlying socket
# for timeout seconds).  See:
# https://2.python-requests.org/en/master/user/quickstart/#timeouts
TIMEOUT = 60

MAX_RETRY_DELAY_SECS = 90

# `request_context` is used to provide information about the request to functions like `download`
# without adding extra function arguments
request_context = {}


def get_retry_delay(retry_num: int, max_delay: int = MAX_RETRY_DELAY_SECS) -> int:
    """The number of seconds to sleep before retrying. Exponential backoff starting
    from 5 seconds up the max_delay. So with a max delay of 60 the retry periods
    would be 5, 10, 20, 40, 60, ..., 60.

    Parameters
        ----------
        retry_num : int
            The current number of times this request has been retried
        max_delay: int
            The maximum number of seconds to wait before retrying
        Returns
        -------
        int : The number of seconds to wait before retrying
    """
    return min(max_delay, 2.5 * (2 ** retry_num))


def is_http(url: str) -> bool:
    """Predicate to determine if the url is an http endpoint.

    Parameters
    ----------
    url : str
        The URL to check

    Returns
    -------
    bool
        Whether the URL is an http endpoint.

    """
    return url is not None and urlparse(url).scheme in ['http', 'https']


def localhost_url(url, local_hostname):
    """Return a version of the url optimized for local development.

    If the url includes the string `localhost`, it will be replaced by
    the `local_hostname`.

    Parameters
    ----------
    url : str
        The url to check
    Returns
    -------
    str : The url, possibly converted to use a different local hostname
    """
    return url.replace('localhost', local_hostname)


def _is_eula_error(body: str) -> bool:
    """
    Tries to determine if the exception is due to a EULA that the user needs to
    approve, and if so, returns a response with the url where they can do so.

    Parameters
    ----------
    body: The body JSON string that may contain the EULA details.

    Returns
    -------
    A boolean indicating if the body contains a EULA error
    """
    try:
        json_object = json.loads(body)
        return "error_description" in json_object and "resolution_url" in json_object
    except Exception:
        return False


def _eula_error_message(body: str) -> str:
    """
    Constructs a user-friendly error indicating the required EULA
    acceptance and the URL where the user can do so.

    Parameters
    ----------
    body: The body JSON string that may contain the EULA details.

    Returns
    -------
    The string with the EULA message
    """
    json_object = json.loads(body)
    return (f"Request could not be completed because you need to agree to the EULA "
            f"at {json_object['resolution_url']}")


@lru_cache(maxsize=128)
def _earthdata_session():
    """Constructs an EarthdataSession for use to download one or more files."""
    return EarthdataSession()


def _add_api_request_uuid(url):
    request_id = request_context.get('request_id')

    if request_id is None:
        return url

    # Parse the URL into components
    parsed_url = urlparse(url)

    # only add the request_id if this is an http/https url
    if parsed_url.scheme != 'http' and parsed_url.scheme != 'https':
        return url

    # Extract the current query parameters from the URL
    query_params = parse_qs(parsed_url.query)

    # Add or update the 'request_id' parameter
    query_params['A-api-request-uuid'] = request_id

    # Convert the query parameters back to a string
    query_string = urlencode(query_params, doseq=True)

    # Rebuild the URL with the new query string
    new_url = urlunparse(
        (parsed_url.scheme, parsed_url.netloc, parsed_url.path,
         parsed_url.params, query_string, parsed_url.fragment)
    )

    return new_url


def _download(
    config, url: str,
    access_token: str,
    data,
    total_retries: int,
    logger, user_agent=None,
    **kwargs_download_agent
):
    """Implements the download functionality.

    Using the EarthdataSession and EarthdataAuth extensions to the
    `requests` module, this function will download the given url.

    Parameters
    ----------
    config : harmony_service_lib.util.Config
        The configuration for the current runtime environment.
    url : str
        The url for the resource to download
    access_token : str
        A shared EDL access token created from the user's access token
        and the app identity.
    data : dict or Tuple[str, str]
        Optional parameter for additional data to send to the server
        when making an HTTP POST request. These data will be URL
        encoded to a query string containing a series of `key=value`
        pairs, separated by ampersands. If None (the default), the
        request will be sent with an HTTP GET request.
    total_retries: int
        Upper limit on the number of times to retry the request
    user_agent : str
        The user agent that is requesting the download.
        E.g. harmony/0.0.0 (harmony-sit) harmony-service-lib/4.0 (gdal-subsetter)
    kwargs_download_agent: dict
        kwargs to be passed to the download agent
        E.g. stream=True

    Returns
    -------
    requests.Response with the download result

    """
    headers = {}
    if user_agent is not None:
        headers['user-agent'] = user_agent
    auth = EarthdataAuth(access_token)
    tries = 0
    retry = True
    response = None
    download_url = None
    while retry is True:
        retry = False
        tries += 1
        try:
            session = _earthdata_session()
            session.auth = auth
            if data is None and len(url) > config.post_url_length:
                parsed_url = urlparse(url)
                data = parsed_url.query
                download_url = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"

            if data is None:
                response = session.get(url, headers=headers, timeout=TIMEOUT, **kwargs_download_agent)
                if response.ok:
                    return response
                else:
                    raise Exception(f'Unable to download due to status code: {response.status_code} \
                        and content {response.content}')
            else:
                # Including this header since the stdlib does by default,
                # but we've switched to `requests` which does not.
                headers['Content-Type'] = 'application/x-www-form-urlencoded'
                response = session.post(
                    download_url if download_url is not None else url,
                    headers=headers,
                    data=data,
                    timeout=TIMEOUT,
                    **kwargs_download_agent)
                if response.ok:
                    return response
                else:
                    raise Exception(f'Unable to download due to status code: {response.status_code} \
                        and content {response.content}')

        except Exception:
            if response is not None and _is_eula_error(response.content):
                msg = _eula_error_message(response.content)
                logger.info(f'{msg} due to: {response.content}')
                return response

            if response is not None and response.status_code in (401, 403):
                msg = f'Forbidden: Unable to download {url}. Will not retry.'
                logger.info(f'{msg} due to: {response.content}')
                return response

            if tries < total_retries:
                retry = True
                delay = get_retry_delay(tries)
                logger.exception(f'Retrying failed download {url}')
                sleep(delay)
            else:
                logger.error(f'All retries exhaused for downloading {url}')
                return response


def _log_download_performance(logger, url, duration_ms, file_size):
    """Logs a message tracking performance information related to a file download.

    Parameters
    ----------
    logger : logging.Logger
        The logger to use.
    url : str
        The url for the resource to download
    duration_ms: int
        The number of milliseconds the download took
    file_size: int
        The size of the downloaded file
    """
    host = 'Unknown'
    url_path = ''
    try:
        match = re.search('.*://([^/]+)(.*)', url)
        if match:
            host = match.group(1)
            url_path = match.group(2)
    except Exception:
        logger.exception(f'Unable to extract host name from {url}')
    extra_fields = {
        'durationMs': duration_ms,
        'host': host,
        "path": url_path,
        "size": file_size
    }
    logger.info('timing.download.end', extra=extra_fields)


def download(config, url: str, access_token: str, data, destination_file,
             user_agent=None, stream=True, buffer_size=1024*1024*16):
    """Downloads the given url using the provided EDL user access token
    and writes it to the provided file-like object.

    Exception cases:
    1. No user access token
    2. Invalid user access token
    3. Unable to authenticate the user with Earthdata Login
       a. User credentials (could happen even after token validation
       b. Application credentials
    4. Error response when downloading
    5. Data requires EULA acceptance by user

    Parameters
    ----------
    config : harmony_service_lib.util.Config
        The configuration for the current runtime environment.
    url : str
        The url for the resource to download
    access_token : str
        A shared EDL access token created from the user's access token
        and the app identity.
    data : dict or Tuple[str, str]
        Optional parameter for additional data to send to the server
        when making an HTTP POST request. These data will be URL
        encoded to a query string containing a series of `key=value`
        pairs, separated by ampersands. If None (the default), the
        request will be sent with an HTTP GET request.
    destination_file : file-like
        The destination file where the data will be written. Must be
        a file-like object opened for binary write.
    user_agent : str
        The user agent that is requesting the download.
        E.g. harmony/0.0.0 (harmony-sit) harmony-service-lib/4.0 (gdal-subsetter)

    Returns
    -------
    requests.Response with the download result

    Side-effects
    ------------
    Will write to provided destination_file
    NOTE: streaming request is used to download the file,
          and the chunksize is defaulted to 16MB based on the experiment with a large file of 1.8Gb
          for optimized speed and memory consumption.
          If you are experiencing some performance decay for high-throughput small-sized granules,
          you may want to set stream=False.
    """

    response = None
    logger = build_logger(config)
    # Add the request ID to the download url so it can be used by Cloud Metrics
    url = _add_api_request_uuid(url)
    start_time = datetime.datetime.now()
    logger.info(f'timing.download.start {url}')

    if (not stream) and buffer_size:
        logger.warn(
            f"In download paramters, buffer_size={buffer_size} will be ignored since stream is set to be {stream}."
        )
    elif stream and not isinstance(buffer_size, int):
        raise Exception(f"In download parameters: buffer_size must be integer when stream={stream}.")

    if access_token is not None:
        response = _download(
            config, url, access_token, data, config.max_download_retries, logger, user_agent, stream=stream
        )

    if response is not None and response.ok:
        if not stream:
            destination_file.write(response.content)
            file_size = sys.getsizeof(response.content)
        else:
            for chunk in response.iter_content(chunk_size=buffer_size):
                destination_file.write(chunk)
            file_size = os.path.getsize(destination_file.name)
        time_diff = datetime.datetime.now() - start_time
        duration_ms = int(round(time_diff.total_seconds() * 1000))
        duration_logger = build_logger(config)
        _log_download_performance(duration_logger, url, duration_ms, file_size)

        return response

    if _is_eula_error(response.content):
        msg = _eula_error_message(response.content)
        logger.info(f'{msg} due to: {response.content}')
        raise ForbiddenException(msg)

    if response.status_code in (401, 403):
        msg = f'Forbidden: Unable to download {url}'
        logger.info(f'{msg} due to: {response.content}')
        raise ForbiddenException(msg)

    msg = f'Unable to download due to status code: {response.status_code} and content \
        {response.content} and all retries exhausted.'
    logger.error(msg)
    raise ServerException(msg)



from unittest import mock
from unittest.mock import Mock, patch

import pytest
import responses

from harmony_service_lib.exceptions import ForbiddenException, ServerException
from harmony_service_lib import util
from tests.util import config_fixture


@pytest.mark.parametrize('url,expected', [
    ('file:///var/log/junk.txt', '/var/log/junk.txt'),
    ('file:///var/logs/virus_scan.txt', '/var/logs/virus_scan.txt'),
    ('/var/logs/game_scores.txt', '/var/logs/game_scores.txt')
])
def test_url_as_filename(url, expected):
    fn = str(util._url_as_filename(url))

    assert fn == expected


def test_when_given_an_s3_uri_it_downloads_the_s3_file(monkeypatch, mocker, faker):
    access_token = faker.password(length=40, special_chars=False)
    aws_download = mocker.Mock()
    monkeypatch.setattr(util.aws, 'download', aws_download)
    config = config_fixture()

    with mock.patch('builtins.open', mock.mock_open()):
        util.download('s3://example/file.txt', 'tmp', access_token=access_token, cfg=config)

    aws_download.assert_called()


def test_when_given_an_http_url_it_downloads_the_url(monkeypatch, mocker, faker):
    access_token = faker.password(length=40, special_chars=False)
    http_download = mocker.Mock()
    monkeypatch.setattr(util.http, 'download', http_download)
    config = config_fixture()

    with mock.patch('builtins.open', mock.mock_open()):
        util.download('https://example.com/file.txt', 'tmp', access_token=access_token, cfg=config)

    http_download.assert_called()


@patch('harmony_service_lib.http.get_retry_delay', Mock(return_value = 0))
def test_when_given_unknown_url_it_raises_exception(faker):
    access_token = faker.password(length=40, special_chars=False)
    config = config_fixture()

    with mock.patch('builtins.open', mock.mock_open()):
        with pytest.raises(Exception):
            util.download('msdos:choplifter.bas', 'tmp', access_token=access_token, cfg=config)


def test_when_given_a_file_url_it_returns_the_file_path(monkeypatch, mocker, faker):
    access_token = faker.password(length=40, special_chars=False)
    http_download = mocker.Mock()
    monkeypatch.setattr(util.http, 'download', http_download)
    config = config_fixture()

    with mock.patch('builtins.open', mock.mock_open()):
        destination_path = util.download('https://example.com/file.txt', '/put/file/here/',
                                         access_token=access_token, cfg=config)

        assert destination_path.startswith('/put/file/here/')
        assert destination_path.endswith('.txt')


def test_when_given_a_file_path_it_returns_the_file_path(monkeypatch, mocker, faker ):
    access_token = faker.password(length=40, special_chars=False)
    http_download = mocker.Mock()
    monkeypatch.setattr(util.http, 'download', http_download)
    config = config_fixture()

    with mock.patch('builtins.open', mock.mock_open()):
        destination_path = util.download('file:///var/logs/example/file.txt', '/put/file/here/',
                                         access_token=access_token, cfg=config)

        assert destination_path.startswith('/var/logs/example/')
        assert destination_path.endswith('.txt')


def test_when_given_file_url_with_parameters_returns_file_path(monkeypatch,
                                                               mocker, faker):
    access_token = faker.password(length=40, special_chars=False)
    http_download = mocker.Mock()
    monkeypatch.setattr(util.http, 'download', http_download)
    config = config_fixture()

    with mock.patch('builtins.open', mock.mock_open()):
        destination_path = util.download(
            'https://example.com/file.nc4?dap4.ce=latitude',
            '/put/file/here/', access_token=access_token, cfg=config
        )

        assert destination_path.startswith('/put/file/here/')
        assert destination_path.endswith('.nc4')


@responses.activate
def test_when_the_url_returns_a_401_it_throws_a_forbidden_exception(faker):
    access_token = faker.password(length=40, special_chars=False)
    url = 'https://example.com/file.txt'
    config = config_fixture()

    responses.add(responses.GET, url, status=401)

    with mock.patch('builtins.open', mock.mock_open()):
        with pytest.raises(ForbiddenException) as e:
            util.download(url, '/tmp', access_token=access_token, cfg=config)
        assert e.value.message.startswith('Forbidden')
        assert len(responses.calls) == 1


@responses.activate
def test_when_the_url_returns_a_403_it_throws_a_forbidden_exception(faker):
    access_token = faker.password(length=41, special_chars=False)
    url = 'https://example.com/file.txt'
    config = config_fixture()
    responses.add(responses.GET, url, status=403)

    with mock.patch('builtins.open', mock.mock_open()):
        with pytest.raises(ForbiddenException) as e:
            util.download(url, '/tmp', access_token=access_token, cfg=config)
        assert e.value.message.startswith('Forbidden')
        assert len(responses.calls) == 1


@responses.activate
def test_when_the_url_returns_a_eula_error_it_returns_a_human_readable_message(faker):
    access_token = faker.password(length=42, special_chars=False)
    url = 'https://example.com/file.txt'
    config = config_fixture()
    responses.add(
        responses.GET, url, status=403,
        body=('{"status_code":403,"error_description":"EULA Acceptance Failure",'
              '"resolution_url":"https://example.com/approve_app?client_id=foo"}')
    )

    with mock.patch('builtins.open', mock.mock_open()):
        with pytest.raises(ForbiddenException) as e:
            util.download(url, '/tmp', access_token=access_token, cfg=config)
        assert e.value.message == (f'Request could not be completed because you need to agree to the EULA '
                                   f'at https://example.com/approve_app?client_id=foo')
        assert len(responses.calls) == 1


@responses.activate
@patch('harmony_service_lib.http.get_retry_delay', Mock(return_value = 0))
def test_when_the_url_returns_a_500_it_does_not_raise_a_forbidden_exception_and_does_not_return_details_to_user(faker):
    access_token = faker.password(length=43, special_chars=False)
    url = 'https://example.com/file.txt'
    config = config_fixture()
    responses.add(
        responses.POST,
        'https://uat.urs.earthdata.nasa.gov/oauth/tokens/user',
        status=200,
        match_querystring=False
    )
    responses.add(responses.GET, url, status=500)

    with mock.patch('builtins.open', mock.mock_open()):
        with pytest.raises(Exception) as e:
            util.download(url, '/tmp', access_token=access_token, cfg=config)
        assert e.type != ForbiddenException and e.type == ServerException
        assert len(responses.calls) == 5



from base64 import b64encode
from dataclasses import dataclass, field

import pytest
from requests import Session

from harmony_service_lib.earthdata import EarthdataAuth, EarthdataSession


@dataclass
class FakeRequest:
    url: str = 'https://fake.download.earthdata.nasa.gov/data'
    headers: dict = field(default_factory=dict)


@pytest.fixture
def earthdata_auth(faker):
    token = faker.password(length=40, special_chars=False)
    return EarthdataAuth(token)


def test_authdata_auth_creates_correct_header(faker):
    token = faker.password(length=40, special_chars=False)
    auth = EarthdataAuth(token)
    request = FakeRequest()

    auth(request)

    assert 'Authorization' in request.headers
    assert 'Bearer' in request.headers['Authorization']
    assert token in request.headers['Authorization']


def test_earthdata_auth_adds_auth_header_(earthdata_auth):
    request = FakeRequest()

    earthdata_auth(request)

    assert 'Authorization' in request.headers

def test_earthdata_auth_removes_auth_header_when_X_Amz_Algorithm_is_set(earthdata_auth):
    request = FakeRequest(url='https://presigned.s3.url.com?X-Amz-Algorithm=foo')

    earthdata_auth(request)

    assert 'Authorization' not in request.headers

def test_earthdata_auth_removes_auth_header_when_signature_is_set(earthdata_auth):
    request = FakeRequest(url='https://presigned.s3.url.com?Signature=bar')

    earthdata_auth(request)

    assert 'Authorization' not in request.headers

def test_earthdata_session_given_no_auth_delegates_to_super(monkeypatch):
    called = False

    def mock_rebuild_auth(self, prepared_request, response):
        nonlocal called
        called = True
    monkeypatch.setattr(Session, 'rebuild_auth', mock_rebuild_auth)
    session = EarthdataSession()

    session.rebuild_auth(None, None)

    assert called


def test_earthdata_session_given_no_auth_header_sets_auth_header(earthdata_auth):
    session = EarthdataSession()
    session.auth = earthdata_auth
    request = FakeRequest()

    session.rebuild_auth(request, None)

    assert 'Authorization' in request.headers


def test_earthdata_session_given_auth_header_replaces_auth_header(earthdata_auth):
    session = EarthdataSession()
    session.auth = earthdata_auth
    request = FakeRequest(headers={'Authorization': 'PreExistingValue'})

    session.rebuild_auth(request, None)

    assert 'Authorization' in request.headers
    assert request.headers['Authorization'] != 'PreExistingValue'



"""
==========
adapter.py
==========

Provides BaseHarmonyAdapter, an abstract base class for services to implement
the translation between Harmony messages and service calls and the translation
between service results and Harmony callbacks.
"""

import logging
import uuid
from abc import ABC
from pystac import Catalog, read_file

from harmony_service_lib.http import request_context
from harmony_service_lib.logging import build_logger
from . import util


class BaseHarmonyAdapter(ABC):
    """
    Abstract base class for Harmony service adapters.  Service implementations
    should inherit from this class and implement the `#invoke(self)` or `#process_item(self, item, source)`
    method to adapt the Harmony message (`self.message`) into a service call

    Services may choose to override methods that do data downloads and result
    staging as well, if they use a different mechanism

    Attributes
    ----------
    message : harmony_service_lib.Message
        The Harmony input which needs acting upon
    logger: Logger
        Logger specific to this request
    """

    def __init__(self, message, catalog=None, config=None):
        """
        Constructs the adapter

        Parameters
        ----------
        message : harmony_service_lib.Message
            The Harmony input which needs acting upon
        catalog : pystac.Catalog
            A STAC catalog containing the files on which to act
        config : harmony_service_lib.util.Config
            The configuration values for this runtime environment.
        """
        # set the request ID in the global context so we can use it in other places
        request_id = message.requestId if hasattr(message, 'requestId') else None
        request_context['request_id'] = request_id

        self.message = message
        self.catalog = catalog
        self.config = config

        if self.config is not None:
            self.init_logging()
        else:
            self.logger = logging.getLogger()

    def set_config(self, config):
        self.config = config
        if self.config is not None:
            self.init_logging()

    def init_logging(self):
        user = self.message.user if hasattr(self.message, 'user') else None
        req_id = self.message.requestId if hasattr(self.message, 'requestId') else None
        logging_context = {
            'user': user,
            'requestId': req_id
        }
        self.logger = logging.LoggerAdapter(build_logger(self.config), logging_context)

    def invoke(self):
        """
        Invokes the service to process `self.message`.  By default, this will call process_item
        on all items in the input catalog

        Returns
        -------
        (harmony_service_lib.Message, pystac.Catalog | list)
            A tuple of the Harmony message, with any processed fields marked as such and
            in this implementation, a single STAC catalog describing the output.
            (Services overriding this method may return a list of STAC catalogs if desired.)
        """
        # New-style processing using STAC
        if self.catalog:
            return (self.message, self._process_catalog_recursive(self.catalog))
        else:
            raise RuntimeError("The service should override the invoke function when no STAC catalog is provided.")

    def get_all_catalog_items(self, catalog: Catalog, follow_page_links=True):
        """
        Returns a lazy sequence of all the items (including from child catalogs) in the catalog.
        Can handle paged catalogs (catalogs with next/prev).

        Parameters
        ----------
        catalog : pystac.Catalog
            The catalog from which to get items
        follow_page_links : boolean
            Whether or not to follow 'next' links - defaults to True

        Returns
        -------
        A generator that can be iterated to provide a lazy sequence of `pystac.Item`s
        """
        # Return immediate items and items from sub-catalogs
        for item in catalog.get_all_items():
            yield item

        # process 'next' link if present
        if follow_page_links:
            link = catalog.get_single_link(rel='next')
            if link:
                next_catalog = read_file(link.get_href())
                next_items = self.get_all_catalog_items(next_catalog, True)
                for item in next_items:
                    yield item

    def _process_catalog_recursive(self, catalog):
        """
        Helper method to recursively process a catalog and all of its children, producing a new
        output catalog of the results

        Parameters
        ----------
        catalog : pystac.Catalog
            The catalog to process

        Returns
        -------
        pystac.Catalog
            A new catalog containing all of the processed results
        """
        result = catalog.clone()
        result.id = str(uuid.uuid4())

        # Recursively process all sub-catalogs
        children = catalog.get_children()
        result.clear_children()
        result.add_children([self._process_catalog_recursive(child) for child in children])

        # Process immediate child items
        items = catalog.get_items()
        item_count = 0
        result.clear_items()
        source = None
        for item in items:
            cloned_item = item.clone()
            # if there is a bbox, but no geometry, create a geometry from the bbox
            if cloned_item.bbox and not cloned_item.geometry:
                cloned_item.geometry = util.bbox_to_geometry(cloned_item.bbox)
            item_count = item_count + 1
            source = source or self._get_item_source(cloned_item)
            output_item = self.process_item(cloned_item, source)
            if output_item:
                # Ensure the item gets a new ID
                if output_item.id == item.id:
                    output_item.id = str(uuid.uuid4())
                result.add_item(output_item)
        self.logger.info(f'Processed {item_count} granule(s)')

        # process 'next' link if present
        link = catalog.get_single_link(rel='next')
        if link:
            next_catalog = read_file(link.get_href())
            result.add_child(self._process_catalog_recursive(next_catalog))

        return result

    def process_item(self, item, source):
        """
        Given a pystac.Item and a message.Source (collection and variables to subset), processes the
        item, returning a new pystac.Item that describes the output location and metadata

        Optional abstract method. Required if the default #invoke implementation is used.  Services
        processing one input file at a time can simplify adapter code by overriding this method.


        Parameters
        ----------
        item : pystac.Item
            the item that should be processed
        source : harmony_service_lib.message.Source
            the input source defining the variables, if any, to subset from the item

        Returns
        -------
        pystac.Item
            a STAC item whose metadata and assets describe the service output
        """
        raise NotImplementedError('subclasses must implement #process_item or override #invoke')

    def _get_item_source(self, item):
        """
        Given a STAC item, finds and returns the item's data source in this.message.  It
        specifically looks for a link with relation "harmony_source" in the item and all
        parent catalogs.  The href on that link is the source collection landing page, which
        can identify a source.  If no relation exists and there is only one source in the
        message, returns the message source.

        Parameters
        ----------
        item : pystac.Item
            the item whose source is needed

        Raises
        ------
        RuntimeError
            if no input source could be unambiguously determined, which indiciates a
            misconfiguration or bad input message

        Returns
        -------
        harmony_service_lib.message.Source
            The source of the input item
        """
        parent = item
        sources = parent.get_links('harmony_source')
        while len(sources) == 0 and parent.get_parent() is not None:
            parent = parent.get_parent()
            sources = parent.get_links('harmony_source')
        if len(sources) == 0:
            if len(self.message.sources) == 1:
                return self.message.sources[0]
            else:
                raise RuntimeError('Could not match STAC catalog to an input source')
        href = sources[0].target
        collection = href.split('/').pop()
        return next(source for source in self.message.sources if source.collection == collection)



"""
Tests STAC-based invocation styles and methods on BaseHarmonyAdapter
"""

import unittest

from unittest.mock import patch
from pystac import Catalog, Item, Link

from harmony_service_lib.message import Message
from harmony_service_lib.adapter import BaseHarmonyAdapter
from .example_messages import full_message
from .util import config_fixture


# Minimal concrete implementation that records method calls
class AdapterTester(BaseHarmonyAdapter):
    process_args = []

    def process_item(self, item, source):
        AdapterTester.process_args.append((item, source))
        if item.id == 'mutate-me':
            item.id = 'i-mutated-you'
        return item


class TestBaseHarmonyAdapterDefaultInvoke(unittest.TestCase):
    def setUp(self):
        AdapterTester.process_args = []
        self.config = config_fixture()

    def test_items_with_no_input_source_raise_exceptions(self):
        catalog = Catalog('0', 'Catalog 0')
        catalog.add_item(Item('1', None, [0, 0, 1, 1], '2020-01-01T00:00:00.000Z', {}))
        adapter = AdapterTester(Message(full_message), catalog, config=self.config)
        self.assertRaises(RuntimeError, adapter.invoke)

    def test_invocation_processes_items_with_sources(self):
        catalog = Catalog('0', 'Catalog 0')
        catalog.add_link(Link('harmony_source', 'http://example.com/C0001-EXAMPLE'))

        message = Message(full_message)
        items = [
            Item('1', None, [0, 0, 1, 1], '2020-01-01T00:00:00.000Z', {}),
            Item('2', None, [0, 0, 1, 2], '2020-01-01T00:00:00.000Z', {})
        ]
        catalog.add_items(items)
        adapter = AdapterTester(message, catalog, config=self.config)
        adapter.invoke()
        self.assertEqual(AdapterTester.process_args[0][0].bbox, items[0].bbox)
        self.assertEqual(AdapterTester.process_args[1][0].bbox, items[1].bbox)
        self.assertEqual(AdapterTester.process_args[0][1], message.sources[0])
        self.assertEqual(AdapterTester.process_args[1][1], message.sources[0])

    @patch('harmony_service_lib.adapter.read_file')
    def test_invocation_follows_linked_catalogs(self, test_patch):
        catalog0 = Catalog('0', 'Catalog 0')
        catalog0.add_link(Link('harmony_source', 'http://example.com/C0001-EXAMPLE'))
        catalog1 = Catalog('1', 'Catalog 1')
        catalog1.add_link(Link('harmony_source', 'http://example.com/C0002-EXAMPLE'))
        catalog0.add_link(Link('next', 'catalog1.json'))
        test_patch.return_value = catalog1
        message = Message(full_message)
        items_a = [
            Item('3', None, [0, 0, 1, 3], '2020-01-01T00:00:00.000Z', {}),
            Item('4', None, [0, 0, 1, 4], '2020-01-01T00:00:00.000Z', {})
        ]
        items_b = [
            Item('5', None, [0, 0, 1, 5], '2020-01-01T00:00:00.000Z', {}),
            Item('6', None, [0, 0, 1, 6], '2020-01-01T00:00:00.000Z', {})
        ]
        catalog0.add_items(items_a)
        catalog1.add_items(items_b)
        adapter = AdapterTester(message, catalog0, config=self.config)
        adapter.invoke()
        self.assertEqual(AdapterTester.process_args[0][0].bbox, items_a[0].bbox)
        self.assertEqual(AdapterTester.process_args[1][0].bbox, items_a[1].bbox)
        self.assertEqual(AdapterTester.process_args[2][0].bbox, items_b[0].bbox)
        self.assertEqual(AdapterTester.process_args[3][0].bbox, items_b[1].bbox)

    def test_invocation_recurses_subcatalogs(self):
        catalog = Catalog('0', 'Catalog 0')
        catalog.add_link(Link('harmony_source', 'http://example.com/C0001-EXAMPLE'))
        catalog.add_child(Catalog('1a', 'Catalog 1a'))
        subcatalog = Catalog('1b', 'Catalog 1b')
        catalog.add_child(subcatalog)
        subsubcatalog_a = Catalog('2a', 'Catalog 2a')
        subsubcatalog_b = Catalog('2b', 'Catalog 2b')
        subsubcatalog_b.add_link(Link('harmony_source', 'http://example.com/C0002-EXAMPLE'))
        subcatalog.add_children([subsubcatalog_a, subsubcatalog_b])

        message = Message(full_message)
        items_a = [
            Item('3', None, [0, 0, 1, 3], '2020-01-01T00:00:00.000Z', {}),
            Item('4', None, [0, 0, 1, 4], '2020-01-01T00:00:00.000Z', {})
        ]
        items_b = [
            Item('5', None, [0, 0, 1, 5], '2020-01-01T00:00:00.000Z', {}),
            Item('6', None, [0, 0, 1, 6], '2020-01-01T00:00:00.000Z', {})
        ]
        subsubcatalog_a.add_items(items_a)
        subsubcatalog_b.add_items(items_b)
        adapter = AdapterTester(message, catalog, config=self.config)
        adapter.invoke()
        self.assertEqual(AdapterTester.process_args[0][0].bbox, items_a[0].bbox)
        self.assertEqual(AdapterTester.process_args[1][0].bbox, items_a[1].bbox)
        self.assertEqual(AdapterTester.process_args[2][0].bbox, items_b[0].bbox)
        self.assertEqual(AdapterTester.process_args[3][0].bbox, items_b[1].bbox)
        self.assertEqual(AdapterTester.process_args[0][1], message.sources[0])
        self.assertEqual(AdapterTester.process_args[1][1], message.sources[0])
        self.assertEqual(AdapterTester.process_args[2][1], message.sources[1])
        self.assertEqual(AdapterTester.process_args[3][1], message.sources[1])

    @patch('harmony_service_lib.adapter.read_file')
    def test_get_all_items_follows_links(self, test_patch):
        catalog0 = Catalog('0', 'Catalog 0')
        catalog0.add_link(Link('harmony_source', 'http://example.com/C0001-EXAMPLE'))
        catalog1 = Catalog('1', 'Catalog 1')
        catalog1.add_link(Link('harmony_source', 'http://example.com/C0002-EXAMPLE'))
        catalog0.add_link(Link('next', 'catalog1.json'))
        catalog1.add_link(Link('prev', 'catalog0.json'))
        test_patch.return_value = catalog1
        message = Message(full_message)
        items_a = [
            Item('3', None, [0, 0, 1, 3], '2020-01-01T00:00:00.000Z', {}),
            Item('4', None, [0, 0, 1, 4], '2020-01-01T00:00:00.000Z', {})
        ]
        items_b = [
            Item('5', None, [0, 0, 1, 5], '2020-01-01T00:00:00.000Z', {}),
            Item('6', None, [0, 0, 1, 6], '2020-01-01T00:00:00.000Z', {})
        ]
        catalog0.add_items(items_a)
        catalog1.add_items(items_b)
        adapter = AdapterTester(message, catalog0, config=self.config)
        default_items = list(adapter.get_all_catalog_items(catalog0))
        self.assertEqual(default_items, [ *items_a, *items_b ])
        all_items = list(adapter.get_all_catalog_items(catalog0, True))
        self.assertEqual(all_items, [ *items_a, *items_b ])
        no_link_items = list(adapter.get_all_catalog_items(catalog0, False))
        self.assertEqual(no_link_items, items_a)

    @patch('harmony_service_lib.adapter.read_file')
    def test_get_all_items_handles_children(self, test_patch):
        catalog = Catalog('0', 'Catalog 0')
        catalog.add_link(Link('harmony_source', 'http://example.com/C0001-EXAMPLE'))
        catalog.add_child(Catalog('1a', 'Catalog 1a'))
        subcatalog = Catalog('1b', 'Catalog 1b')
        catalog.add_child(subcatalog)
        subsubcatalog_a = Catalog('2a', 'Catalog 2a')
        subsubcatalog_b = Catalog('2b', 'Catalog 2b')
        subsubcatalog_b.add_link(Link('harmony_source', 'http://example.com/C0002-EXAMPLE'))
        subcatalog.add_children([subsubcatalog_a, subsubcatalog_b])

        message = Message(full_message)
        items_a = [
            Item('3', None, [0, 0, 1, 3], '2020-01-01T00:00:00.000Z', {}),
            Item('4', None, [0, 0, 1, 4], '2020-01-01T00:00:00.000Z', {})
        ]
        items_b = [
            Item('5', None, [0, 0, 1, 5], '2020-01-01T00:00:00.000Z', {}),
            Item('6', None, [0, 0, 1, 6], '2020-01-01T00:00:00.000Z', {})
        ]
        subsubcatalog_a.add_items(items_a)
        subsubcatalog_b.add_items(items_b)
        adapter = AdapterTester(message, catalog, config=self.config)
        all_items = list(adapter.get_all_catalog_items(catalog, True))
        self.assertEqual(all_items, [ *items_a, *items_b ])

    def test_unaltered_ids_are_assigned_new_uuids(self):
        catalog = Catalog('0', 'Catalog 0')
        catalog.add_link(Link('harmony_source', 'http://example.com/C0001-EXAMPLE'))

        message = Message(full_message)
        items = [
            Item('1', None, [0, 0, 1, 1], '2020-01-01T00:00:00.000Z', {}),
            Item('2', None, [0, 0, 1, 1], '2020-01-01T00:00:00.000Z', {})
        ]
        catalog.add_items(items)
        adapter = AdapterTester(message, catalog, config=self.config)
        (message, out_catalog) = adapter.invoke()
        self.assertNotEqual(out_catalog.id, catalog.id)

        out_items = [item for item in out_catalog.get_items()]
        self.assertNotEqual(out_items[0].id, items[0].id)
        self.assertNotEqual(out_items[1].id, items[1].id)

    def test_altered_ids_are_retained(self):
        catalog = Catalog('0', 'Catalog 0')
        catalog.add_link(Link('harmony_source', 'http://example.com/C0001-EXAMPLE'))

        message = Message(full_message)
        items = [
            Item('mutate-me', None, [0, 0, 1, 1], '2020-01-01T00:00:00.000Z', {}),
            Item('2', None, [0, 0, 1, 1], '2020-01-01T00:00:00.000Z', {})
        ]
        catalog.add_items(items)
        adapter = AdapterTester(message, catalog, config=self.config)
        (message, out_catalog) = adapter.invoke()
        out_items = [item for item in out_catalog.get_items()]
        self.assertEqual(out_items[0].id, 'i-mutated-you')



from unittest import TestCase

from freezegun import freeze_time

from harmony_service_lib.provenance import get_updated_history_metadata


@freeze_time('2001-02-03T04:05:06')
class TestProvenance(TestCase):
    """Test class for functions in harmony_service_lib.provenance."""

    def test_get_updated_history_metadata_existing_history(self):
        """A new line should be appended to the existing history metadata."""
        self.assertEqual(
            get_updated_history_metadata(
                'Amazing Service',
                '1.2.3',
                '1999-12-11T10:09:08 File created',
            ),
            '1999-12-11T10:09:08 File created\n2001-02-03T04:05:06+00:00 Amazing Service 1.2.3',
        )

    def test_get_updated_history_metadata_no_existing_history(self):
        """The output should only be the newly created history metadata."""
        self.assertEqual(
            get_updated_history_metadata(
                'Amazing Service',
                '1.2.3',
            ),
            '2001-02-03T04:05:06+00:00 Amazing Service 1.2.3',
        )
